#! /usr/bin/perl -w
#
# AUTHOR
#    Alexis Domjan <adomjan@horus.ch>
# LICENSE
#    Copyright (C) 2005, under GNU Public LIcense
#
# $Id: elizabot.pl,v 1.22 2005/01/25 17:29:54 disco Exp $
#

use Net::IRC;
use Chatbot::Eliza;
use Time::Local;

# We use the database
use lib '/usr/lib/gppds';
use gppds;

# IRC/Database handlers
my $irc;
my $conn;
my $dbh;

# Hashes
my %config;
my %notes;
my %auth;
my %last_sentences;
my %cmdusage;

# Vars
my $chatbot;
my $realname;
my $realnick;
my $global_userhost;

# Wait Time
my $wait_time = 0.5;

# Definitions of Eliza available commands

my %eliza_cmds = (	'test'		=> {	name => 'test',
						help => 'This is a test command',
						response => 'public',
						external => undef,
						internal => \&internal_command_test },
	
			'help'		=> {	name => 'help',
						help => 'This help',
						response => 'private',
						external => undef,
						internal => \&internal_command_help },

			'cff'		=> {	name => 'cff',
						help => 'CFF timetable, usage: !cff "source" "dest" [mm.dd.yyyy] [hh:mm]',
						response => 'public',
						external => 'external/cff',
						internal => undef },

			'solarwind'	=> {	name => 'solarwind',
						help => 'Give some informations about solar wind.',
						response => 'public',
						external => 'external/spaceweather',
						internal => undef },

			'note'		=> {	name => 'note',
						help => 'Notes service, usage: !note <destination> <message>',
						response => 'public',
						external => undef,
						internal => \&internal_command_note },

			'auth'		=> {	name => 'auth',
						help => 'Authentify yourself with your database password! WARNING: Only use by sending a private message to the bot! usage: !auth <nick> <password>',
						response => 'private',
						external => undef,
						internal => \&internal_command_auth },

			'currency'	=> { 	name => 'currency',
						help => 'Currency converter: example: !currency CHF EUR 23. You can use the \$CHFEUR:23 shortcut.',
						response => 'public',
						external => 'external/currenciesconvert',
						internal => undef },

			'whoami'	=> { 	name => 'whoami',
						help => 'Gnothi Seauton',
						response => 'public',
						internal => \&internal_command_whoami,
						external => undef },

			'smsquota'	=> { 	name => 'smsquota',
						help => '[AUTH] Return your quota of SMS.',
						response => 'public',
						internal => \&internal_command_smsquota,
						external => undef },

			'sms'		=> {	name => 'sms',
						help => '[AUTH] Send an SMS. usage: !sms <recipient> <message>',
						response => 'private',
						internal => \&internal_command_sms,
						external => undef },

			'date'		=> {	name => 'date',
						help => 'Return current local date.',
						response => 'public',
						internal => \&internal_command_date,
						external => undef },

			'trans'		=> {	name => 'trans',
						help => 'Translation system. usage: ?xy:text, where x is the source language, y the destination and text the text to translate.',
						response => 'public',
						internal => undef,
						external => 'external/googletrans',
						external_prefix => '-> ' },

			'ask'		=> {	name => 'ask',
						help => 'Ask questions to Eliza',
						response => 'public',
						internal => \&internal_command_ask },

			'fortune'	=> {	name => 'fortune',
						help => 'Famous fortune',
						response => 'public',
						external => '/usr/games/fortune' },

			'excuse'	=> {	name => 'excuse',
						help => 'Famous BOFH excuse',
						response => 'public',
						external => '/usr/games/fortune bofh-excuses' },

			'ping'		=> {	name => 'ping',
						help => 'Perform an ICMP ping to the host specified as argument',
						response => 'public',
						internal => \&internal_command_ping },

			'till'		=> {	name => 'till',
						help => 'Count time until given date (supports since). usage: !till YYYY.MM.DD hh:mm',
						response => 'public',
						internal => \&internal_command_till },

			'since'		=> {	name => 'since',
						help => 'Count time since given date (supports till). usage: !since YYYY.MM.DD hh:mm',
						response => 'public',
						internal => \&internal_command_till },

			'birth'		=> {	name => 'birth',
						help => 'Statistics with your birthdate. usage: !birth YYYY.MM.DD hh:mm',
						response => 'public',
						internal => \&internal_command_birth },

			'mail'		=> {	name => 'mail',
						help => 'Send an email. usage: !mail <recipient> (<subject>) <message>',
						response => 'public',
						internal => \&internal_command_mail },

			'tcping'	=> {	name => 'tcping',
						help => 'Perform a TCP ping. usage: !tcping <host> <port>',
						response => 'public',
						internal => \&internal_command_tcping },

			'cfm'		=> {	name => 'cfm',
						help => 'Stands for Choose For Me. !cfm <choice1>,<choice2>,...,<choicen>',
						response => 'public',
						internal => \&internal_command_cfm },

			'info'		=> {	name => 'info',
						help => 'Get informations about a nickname. usage: !info <nick> <info>. Valid info are: nick, name, birthdate, location, profession, sex, gsm, homepage, and some more ;)',
						response => 'public',
						internal => \&internal_command_info },

			'seen'		=> {	name => 'seen',
						help => 'Display when someone has been seen on the channel. usage: !seen <nickname>',
						response => 'public',
						internal => \&internal_command_seen },

			'calc'		=> {	name => 'calc',
						help => 'Calculator. usage: =<expr>, or !calc <expr>',
						response => 'public',
						internal => \&internal_command_calc },
							
			'dc'		=> {	name => 'dc',
						help => 'Polish Calculator. usage: !dc <expr>, for example: !dc 3 4 5 + * p',
						response => 'public',
						internal => \&internal_command_dc },
							

		);







# Definitions of internal commands

sub internal_command_test {
   my ($nick, $userhost, $params)  = @_;
   return "$nick: this is just a test";
}

sub internal_command_help {
   my ($nick, $userhost, $params)  = @_;
   my $lines = "";
   my $n;

   if ($params) {
      if ($eliza_cmds{$params}) {
         $lines = $params . ": " . $eliza_cmds{$params}->{help} . "\n";
      } else {
         $lines = "This command is not available!";
      }
   } else {
      #$lines = "*** This is a list of Eliza's Commands ***\n";
      #foreach my $i (keys %eliza_cmds) {
      #   my $numsp = 12 - length($i);
      #   if ($numsp < 1) { $numsp = 1; }
      #   $lines .= $i . " "x$numsp . " : " . $eliza_cmds{$i}->{help} . "\n";
      #}
      $lines  = "*** List of Eliza's Commands. For more information type !help <command>\n";
      if ($realnick ne "unknown_nick") {
         $lines .= "*** in parenthesis, number of time you used the command\n";
         foreach my $cmd (sort {$a cmp $b} keys %eliza_cmds) {
            if ($cmdusage{$realnick}->{$cmd}) {
               $n = $cmdusage{$realnick}->{$cmd};
            } else { 
               $n = 0;
            }
            $lines .= $cmd . "($n), ";
         }
         $lines = substr($lines, 0, length($lines) - 1) # remove trailing comma
      } else {
         $lines .= join(', ', sort {$a cmp $b} keys %eliza_cmds);
      }
        
   }
   return $lines;
}



sub internal_command_whoami {
   my ($nick, $userhost, $params) = @_;
   if ($realnick eq "unknown_nick") {
      return "$nick: Sorry, I don't know you. You may consider to add this nick ($nick) as an alias.";
   } else {
      if (&verify_auth( $realnick, $nick, $userhost)) {
         return "$nick: You're known as $realnick and your name is $realname and you're authenticated!";
      } else {
         return "$nick: You're known as $realnick and your name is $realname. You're not authenticated!";
      }
   }
}



sub internal_command_note {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^(.*?) (.*?)$/) {
      my $recip = $1;
      my $msg = $2;
      my $date = `date`;
      chomp $date;

      # check if recipient exists
      my $realrecip = &getinfo_from_alias( $recip, 'nick' ) or return "$nick: Please use an existing nick or alias as recipient";
      my $realfrom = &getinfo_from_alias( $nick, 'nick' ) || $nick;
      $notes{$realrecip}->{$date}->{from} = $nick;
      if (&verify_auth($realfrom, $nick, $userhost)) {
         $notes{$realrecip}->{$date}->{message} = $msg . " [authenticated]";
      } else {
         $notes{$realrecip}->{$date}->{message} = $msg; 
      }
         
      return "$nick: Your note as been stored for $recip (as $realrecip)!";
      
   } else {
      return "$nick: You must provide recipient nick and message";
   }
}



sub internal_command_auth {
   my ($nick, $userhost, $params) = @_;

   # Authentification is stored in a hash which contain :
   # 1. the real nickname of the user bound in the db
   # 2. the followed nick used on the channel, this may change if user use /nick
   # 3. the hostname of the user, beeing checked
   # 4. the epoch, just in case we want to add some kind of timeout
   #
   print "DEBUG: auth command launched for $nick, $userhost\n";

   my ($auth_nick, $auth_password) = split /\s+/, $params;
   
   # Check for nick usurpation
   if ($realnick ne $auth_nick) {
      return "$nick: don't try to authenticate as someone else than you!";
   }
   
   if (($auth_nick) && ($auth_password)) {

      # Get user hashed password
      my $password = &getinfo_from_alias( $auth_nick, 'password' );
      my $salt = "";
      if ($password =~ /^(..).*?$/) {
         $salt = $1;
      }

      # compare passwords
      if ( crypt($auth_password, $salt) eq $password ) {
         # set authentication
         $auth{$realnick}->{nick} = $nick;
	 $auth{$realnick}->{userhost} = $userhost;
	 $auth{$realnick}->{epoch} = time;
         return "$nick: you're authenticated successfully as $realnick from $userhost.";
      } else {
         return "$nick: wrong nickname or password";
      }
  } else {
     return "$nick: you must provide nickname and password";
  }

}


sub internal_command_smsquota {
   my ($nick, $userhost, $params)  = @_;
   
   # verify authentication
   if (&verify_auth($realnick, $nick, $userhost)) {
      return "$nick: Your quota is " . &getsmsquota($nick) . " credits.";
   } else {
      return "$nick: You've to authenticate yourself first!";
   }
}

   
sub internal_command_sms {
   my ($nick, $userhost, $params)  = @_;

   my $number = "";
   my $message = "";
   my $alias = "";
   
   # verify authentication
   if (&verify_auth($realnick, $nick, $userhost)) {
      # We can send the SMS
      # First let's check the recipient number, if it's a number
      if ($params =~ /^(\d+)\s(.*?)$/) {
         $number = $1;
         $message = $2;
         $alias = $number;
      } elsif ($params =~ /^(\S+)\s(.*?)$/) {
         $alias = $1;
         if ($number = &get_gsm_number_from_phonebook( $realnick, $alias ) ) {
            $message = $2;
         } else {
            return "$nick: Alias '$alias' not found in your SMS phonebook.";
         }
      }

      # Get the *from* number
      my $from = &getinfo_from_alias( $realnick, 'gsm' );

      my $quota = &getsmsquota( $realnick );
      my $count = &getsmscount( $realnick );

      return "$nick: Sorry, you've not enough quota ($quota) to send SMS!" if $quota < 1;

      # We can send the SMS now
      print "Sending to $number from $from\n";
      $from =~ s/\+/00/;
      $number =~ s/\+/00/;

      
      # sending SMS
      my $used_credits = `echo \"$message\" | /usr/bin/send_sms.pl ~/.smssendrc $from $number`;
      chomp $used_credits;

      if ($used_credits) {
         my $new_credit = $quota - $used_credits;
         my $new_count = &getsmscount($realnick) + $used_credits;
         # update credit
         if (!&do_query(     $dbh,
                             'UPDATE sms_quota SET count=?,credit=? WHERE user_id=(SELECT id FROM users WHERE nick=?)',
                             undef,
                             undef,
                             [ $new_count, $new_credit, $realnick ],
                             undef,
                             $error_reason ) ) {
                     print "sql error: $error_reason\n";
         } else {
         }
      }

      return "$nick: SMS has been sent to $alias and it used $used_credits credits!";

   } else {
      return "$nick: You've to authenticate yourself first!";
   }
}



sub internal_command_date {
   my ($nick, $userhost, $params) = @_;

   if ($params) {
      $ENV{TZ} = $params;
      my $ret = "Current time in $params zone is: " . scalar(localtime());
      $ENV{TZ} = "Europe/Zurich";
      return $ret;
   } else {
      my $ret = "Current time in Switzerland is " . scalar localtime(time);
      my ($tsec, $tmin, $thour, @foo) = gmtime(time);
      $thour = $thour + 1;
      $thour = 0 if ($thour == 24);
      my $beat = int(1000 * ($tsec/60 + $tmin + $thour * 60) / 1440);
      $ret .= ". And in the whole world, the internet time is \@$beat\n";
      return $ret;
   }
}


sub internal_command_ask {
   my ($nick, $userhost, $params) = @_;

   if (!($chatbot)) {
      $chatbot = new Chatbot::Eliza;
   } 

   my $answer = $chatbot->transform($params);

   return "$nick: $answer";
}


sub internal_command_ping {
   my ($nick, $userhost, $params) = @_;

   my $hostname;

   if ($params =~ /^([-.A-Za-z0-9]+)$/) {
      my $hostname = $1;
      if (open HOST, "host $hostname 2>&1 |") {
         my $line = <HOST>;
         chomp $line;
         close HOST;

         if ($line =~ /not found/) {
            return "$nick: host $hostname does not exist!";
         }

         if (open PING, "/bin/ping -c 1 $hostname |") {
            my $line = <PING>;
	    chomp $line;
	    $line = <PING> if $line eq "";
	    my ($foo, $host, $ip, $bar) = split / /, $line;
	    $ip =~ s/\:$//;
	    my $line2 = <PING>;
	    if ($line2 =~ /from /) {
	       return "$nick: Host $hostname $ip responded to an ICMP ping and is alive.";
	    } else {
	       return "$nick: Host $hostname $ip didn't respond to an ICMP ping, could be dead or filtering ICMP, use !tcping instead.";
	    }
	 } else {
	    return "$nick: can't run /bin/ping!";
	 }

         
      } else {
         return "$nick: 'host' command not found on local system";
      }
   } else {
      return "$nick: Bad hostname/IP";
   }
}




sub internal_command_till {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^(\d+).(\d+).(\d+)\s+(\d+):(\d+)$/) {
      my $bday = $1;
      my $bmon = $2;
      my $byear = $3;
      my $bhour = $4;
      my $bmin = $5;
      
      if ( ($bday  > 31) || ($bday  < 1) ) { return "$nick: Invalid day!"; }
      if ( ($bmon  > 12) || ($bmon  < 1) ) { return "$nick: Invalid month!"; }
      if ( ($bhour > 23) || ($bhour < 0) ) { return "$nick: Invalid hour!"; }
      if ( ($bmin  > 59) || ($bmin  < 0) ) { return "$nick: Invalid minute!"; }
      if ($byear < 1970) { return "$nick: Invalid year, should be >= 1970 and <2038"; }
   
      $bmon--;
      my $time = time;
      my $nowyear = (localtime)[5];
      my $birth = timelocal(0,$bmin,$bhour,$bday,$bmon,$byear);
      my $target;
      #if($action eq 'birth') {
      #        $target = timelocal(0,$bmin,$bhour,$bday,$bmon,$nowyear);
      #        $target = timelocal(0,$bmin,$bhour,$bday,$bmon,$nowyear+1) if $target < $time;
      #} else {
      $target = timelocal(0,$bmin,$bhour,$bday,$bmon,$byear);
      #}
      $nowyear += 1900 if ($nowyear>=70);
      $nowyear += 2000 if ($nowyear<70);
      my $now   = time;
      my $diff  = $now - $birth;
      my $day   = int($diff / 86400);
      my $hour  = int( ($diff - $day * 86400) / 3600 );
      my $min   = int( ($diff - $day * 86400 - $hour * 3600) / 60 );
      my $sec   = int( ($diff - $day * 86400 - $hour * 3600 - $min * 60) );
      #IRC::send_raw "PRIVMSG $channel :" . "You were born on " . scalar(localtime($birth)) . ". You're $day days, $hour hour" . "s"x($hour > 1) . ", $min minute" . "s"x($min > 1) . " and $sec second" . "s"x($sec > 1) . " old. Your heart probably beat about " . int($diff * (70/60)) . " times :-)\n" if $action eq 'birth';
      $diff  = $target - $now;
      $day   = int($diff / 86400);
      $hour  = int( ($diff - $day * 86400) / 3600 );
      $min   = int( ($diff - $day * 86400 - $hour * 3600) / 60 );
      $sec   = int( ($diff - $day * 86400 - $hour * 3600 - $min * 60) );
      # my $desc = $action eq 'birth' ? "And your next birthday" : "That";
      #IRC::send_raw "PRIVMSG $channel :$desc will be in $day day" . "s"x($day != 1) .  ", $hour hour" . "s"x($hour != 1) . ", $min minute" . "s"x($min != 1) . " and $sec second" . "s"x($sec != 1) . ".\n" if $diff > 0;
      return "$nick: That will be in $day day" . "s"x($day != 1) .  ", $hour hour" . "s"x($hour != 1) . ", $min minute" . "s"x($min != 1) . " and $sec second" . "s"x($sec != 1)  if $diff > 0;
      $day *= -1 if $day;
      $hour *= -1 if $hour;
      $min *= -1 if $min;
      $sec *= -1 if $sec;
      return "$nick: That was $day day" . "s"x($day != 1) . ", $hour hour" . "s"x($hour != 1) . ", $min minute" . "s"x($min != 1) . " and $sec second" . "s"x($sec != 1) . " ago.\n" if $diff < 0;
      return "$nick: That is right now!\n" unless $diff;

   } else {
      return "$nick: Use 'DD.MM.YYYY hh:mm' format.";
   }
}




sub internal_command_birth {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^(\d+).(\d+).(\d+)\s+(\d+):(\d+)$/) {
      my $bday = $1;
      my $bmon = $2;
      my $byear = $3;
      my $bhour = $4;
      my $bmin = $5;
      
      if ( ($bday  > 31) || ($bday  < 1) ) { return "$nick: Invalid day!"; }
      if ( ($bmon  > 12) || ($bmon  < 1) ) { return "$nick: Invalid month!"; }
      if ( ($bhour > 23) || ($bhour < 0) ) { return "$nick: Invalid hour!"; }
      if ( ($bmin  > 59) || ($bmin  < 0) ) { return "$nick: Invalid minute!"; }
      if ($byear < 1970) { return "$nick: Invalid year, should be >= 1970 and <2038"; }
   
      $bmon--;
      my $time = time;
      my $nowyear = (localtime)[5];
      my $birth = timelocal(0,$bmin,$bhour,$bday,$bmon,$byear);
      my $target;
      $target = timelocal(0,$bmin,$bhour,$bday,$bmon,$nowyear);
      $target = timelocal(0,$bmin,$bhour,$bday,$bmon,$nowyear+1) if $target < $time;
      $nowyear += 1900 if ($nowyear>=70);
      $nowyear += 2000 if ($nowyear<70);
      my $now   = time;
      my $diff  = $now - $birth;
      my $day   = int($diff / 86400);
      my $hour  = int( ($diff - $day * 86400) / 3600 );
      my $min   = int( ($diff - $day * 86400 - $hour * 3600) / 60 );
      my $sec   = int( ($diff - $day * 86400 - $hour * 3600 - $min * 60) );
      my $ret = "$nick: You were born on " . scalar(localtime($birth)) . ". You're $day days, $hour hour" . "s"x($hour > 1) . ", $min minute" . "s"x($min > 1) . " and $sec second" . "s"x($sec > 1) . " old. Your heart probably beat about " . int($diff * (70/60)) . " times :-)\n";
      $diff  = $target - $now;
      $day   = int($diff / 86400);
      $hour  = int( ($diff - $day * 86400) / 3600 );
      $min   = int( ($diff - $day * 86400 - $hour * 3600) / 60 );
      $sec   = int( ($diff - $day * 86400 - $hour * 3600 - $min * 60) );
      $ret .= "$nick: And your next birthday will be in $day day" . "s"x($day != 1) .  ", $hour hour" . "s"x($hour != 1) . ", $min minute" . "s"x($min != 1) . " and $sec second" . "s"x($sec != 1) . ".\n" if $diff > 0;
      $day *= -1 if $day;
      $hour *= -1 if $hour;
      $min *= -1 if $min;
      $sec *= -1 if $sec;
      return $ret;

   } else {
      return "$nick: Use 'DD.MM.YYYY hh:mm' format.";
   }
}



sub internal_command_mail {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^([0-9A-Za-z\@-_\.]+)\s+\((.*?)\)\s+(.*?)$/) {
      my $recipient = $1;
      my $subject = $2;
      my $text = $3;

      my $authentic = undef;

      if (&verify_auth($realnick, $nick, $userhost)) {
         $authentic = 1;
      }

      open TMP, "| /usr/sbin/sendmail -i $recipient" or return "$nick: no /usr/sbin/sendmail on this machine!";
      print TMP "From: $config{return_email_address}\n";
      print TMP "Subject: $subject\n";
      print TMP "\n";
      print TMP "Hello, you got the following message from $nick ($realname, known as $realnick) on channel $config{channel} ";
      if ($authentic) {
         print TMP "(Sender has been authenticated by password) :\n" if $authentic;
      } else {
         print TMP "(Not authenticated!) :\n" if !$authentic;
      }
      print TMP "\n";
      print TMP "$text\n";
      close TMP;

      return "$nick: Your message has been sent to $recipient.";
      
   } else {
      return "$nick: Wrong syntax, use !mail <recipient> (<subject>) <message>";
   }
}



sub internal_command_cfm {
   my ($nick, $userhost, $params) = @_;

   my @chces = split /,/, $params;

   my $result = $chces[rand(scalar @chces)];

   return "$nick: $result";
}




sub internal_command_tcping {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^([A-Za-z0-9\.-_]+)\s([a-z0-9-]+)$/) {
      my $host = $1;
      my $port = $2;

      if (system('nc', '-z', '-w1', $host, $port) == 256) {
         return "$nick: Host $host:$port is dead";
      } else {
         return "$nick: Host $host:$port is alive";
      }
   } else {
      return "$nick: Wrong syntax. Usage: !tcping <host> <port>";
   }
}


sub internal_command_info {
   my ($nick, $userhost, $params) = @_;

   if ($params =~ /^([A-Za-z0-9_\-\.]+)\s([a-z]+)$/) {
      my $info = $2;
      my $nickname = $1;
      if ( ($info eq "password") || ($info eq "inlove") ) {
         return "$nick: This kind of information cannot be requested.";
      } else {
         my $res;
         if ($res = &getinfo_from_alias( $nickname, $info )) {
            if ($info eq "gsm") {
               $res =~ s/(..)$/xx/;
            }
	    return "$nick: $nickname ($info) : $res";
	 } else {
	    return "$nick: Invalid information or bad nickname. usage: !info <nick> <info>.";
	 } 
      }
   } else {
      return "$nick: Wrong syntax: usage: !info <nick> <info>.";
   }
}



sub internal_command_seen {
   my ($nick, $userhost, $params) = @_;
   
   if ($params =~ /^([A-Za-z0-9]+)$/) {
      my $nickname = $1;

      my $lastseen = &getinfotoday_from_alias( $nickname, 'lastseen' );

      return "Last time $nickname did something here was on (5 minutes wrap) : $lastseen";
   } else {
      return "$nick: Wrong syntax: usage: !seen <nick>.";
   }
}






sub internal_command_calc {
   my ($nick, $userhost, $params) = @_;

   my $expr = $params;
   $expr =~ s/\\n/\n/g;

   open BC, "| bc -l > /tmp/eliza-bc-result" or return "$nick: Unable to start UNIX calculator (bc)!";
   print BC $expr . "\n";
   close BC;

   my $res = "";
   open BC, "/tmp/eliza-bc-result" or return "$nick: Unable to read result!";
   while (my $line = <BC>) {
      chomp $line;
      $line =~ s/\\//g;
      $res .= $line;
   }
   close BC;

   if (length($res) > 500) {
      $res = substr($res, 1, 500);
      $res .= " [truncated]";
   }

   return "$expr = $res";
}



sub internal_command_dc {
   my ($nick, $userhost, $params) = @_;

   my $expr = $params;

   open DC, "| dc > /tmp/eliza-dc-result" or return "$nick: Unable to start UNIX Arbitrary calculator (dc)!";
   print DC $expr . "\n";
   close DC;

   my $res = "";
   open DC, "/tmp/eliza-dc-result" or return "$nick: Unable to read result!";
   while (my $line = <DC>) {
      chomp $line;
      $line =~ s/\\//g;
      $res .= $line;
   }
   close DC;

   if (length($res) > 500) {
      $res = substr($res, 1, 500);
      $res .= " [truncated]";
   }

   return "=> $res";
}



   


# Functions

# NAME
#    read_config 
sub read_config {
    my ( $file ) = @_;
    my %config;
    open( FILE, $file ) or return 0;
    while (<FILE>) {
        chomp;
        next if /^#/; # comment
        next if /^$/; # empty line
        my ( $name, $value ) = split ( /=/, $_, 2 );
        $config{$name} = $value;
    }
    close(FILE);
    return %config;
}



# NAME
#    write_config
sub write_config {
    my ( $file, $ref ) = @_;
    open( FILE, ">$file" ) or return 0;
    for ( keys %{$ref} ) {
        print FILE "$_=$ref->{$_}\n";
    }
    close(FILE);
}




# NAME
#    init
sub init {
   %config = &read_config( 'elizabot.conf' );

   # Set up the connection to the IRC server.
   $irc = new Net::IRC;
   $conn = $irc->newconn(	Server  => $config{server},
				Port	=> $config{port},
				Nick	=> $config{nick},
				Ircname => $config{ircname} );

   # Connect the handlers to the events.
   $conn->add_handler( 376, \&join_channel ); # end of /motd
   $conn->add_handler( 422, \&join_channel ); # no motd
   $conn->add_handler( 'public', \&message );
   $conn->add_handler( 'msg', \&message );
   $conn->add_handler( 'nick', \&nickchange );
   $conn->add_handler( 'part', \&part );
   $conn->add_handler( 'quit', \&quit );
   $conn->add_handler( 'caction', \&action );
   
   # If we got an alarm, then we read the external file.
   # We could configure an alarm timeout to read every
   # minute internally
   $SIG{ALRM} = \&read_comm_file;
   alarm 0;


   # Opening database
   if ($dbh = &open_database( $config{db_name}, $config{db_host}, $config{db_user}, $config{db_password} ) ) {
      print "database connected!\n";
   } else {
      print "cannot connect database...\n";
   }

   # 
   $| = 1;

   # STARTS irc
   $irc->start();


}

# NAME
#    verify_auth
sub verify_auth {
   my ($real_nickname, $used_nickname, $actual_userhost) =@_;

   if ($auth{$real_nickname}) {
      if ( ($auth{$real_nickname}->{nick} eq $used_nickname) &&
           ($auth{$real_nickname}->{userhost} eq $actual_userhost) ) {
         return 1;
      } else {
         return undef;
      }
   } else {
      return undef;
   }
}
      


# NAME
#    getinfo_from_alias
sub getinfo_from_alias {
        my ($alias, $what) = @_;
        my @titles;
        my @content;
        my $error_reason;
        my %hash;
        if (!&do_query( $dbh,
                        "SELECT u.$what FROM users u, alias a WHERE u.id = a.nick_ref AND a.alias = ?",
                        \@titles,
                        \@content,
                        [ $alias ],
                        undef,
                        \$error_reason ) ) {
                        # error
                        print STDERR "ERROR: $error_reason\n";
        } else {
                        if (scalar(@content) > 0) {
                                %hash = %{$content[0]};
                        }
        }
        if (defined($hash{$what})) {
                return $hash{$what};
        } else {
                return undef;
        }
}



# NAME
#    getinfo_from_alias
sub getinfotoday_from_alias {
        my ($alias, $what) = @_;
        my @titles;
        my @content;
        my $error_reason;
        my %hash;
        if (!&do_query( $dbh,
                        "SELECT u.$what FROM users_today u, alias a WHERE u.id = a.nick_ref AND a.alias = ?",
                        \@titles,
                        \@content,
                        [ $alias ],
                        undef,
                        \$error_reason ) ) {
                        # error
                        print STDERR "ERROR: $error_reason\n";
        } else {
                        if (scalar(@content) > 0) {
                                %hash = %{$content[0]};
                        }
        }
        if (defined($hash{$what})) {
                return $hash{$what};
        } else {
                return undef;
        }
}




# NAME
#    getsmsquota
sub getsmsquota {
        my ($alias) = @_;
        my @titles;
        my @content;
        my $error_reason;
        my %hash;
        if (!&do_query( $dbh,
                        "SELECT sq.credit, sq.count FROM sms_quota sq, users u, alias a WHERE u.id = a.nick_ref AND a.alias = ? AND sq.user_id = u.id",
                        \@titles,
                        \@content,
                        [ $alias ],
                        undef,
                        \$error_reason ) ) {
                        # error
                        print STDERR "ERROR: $error_reason\n";
        } else {
                        if (scalar(@content) > 0) {
                                %hash = %{$content[0]};
                        }
        }
        if (defined($hash{credit})) {
                return $hash{credit};
        } else {
                return undef;
        }
}


# NAME
#    getsmscount
sub getsmscount {
        my ($alias) = @_;
        my @titles;
        my @content;
        my $error_reason;
        my %hash;
        if (!&do_query( $dbh,
                        "SELECT sq.credit, sq.count FROM sms_quota sq, users u, alias a WHERE u.id = a.nick_ref AND a.alias = ? AND sq.user_id = u.id",
                        \@titles,
                        \@content,
                        [ $alias ],
                        undef,
                        \$error_reason ) ) {
                        # error
                        print STDERR "ERROR: $error_reason\n";
        } else {
                        if (scalar(@content) > 0) {
                                %hash = %{$content[0]};
                        }
        }
        if (defined($hash{count})) {
                return $hash{count};
        } else {
                return undef;
        }
}




# NAME
#    get_gsm_number_from_phonebook
sub get_gsm_number_from_phonebook {
        my ($real_nickname, $alias) = @_;
        my @titles;
        my @content;
        my $error_reason;
        my %hash;
        if (!&do_query( $dbh,
                        "SELECT pb.gsm FROM phonebook pb, users u WHERE u.nick = ? AND pb.user_id = u.id AND pb.name = ?",
                        \@titles,
                        \@content,
                        [ $real_nickname, $alias ],
                        undef,
                        \$error_reason ) ) {
                        # error
                        print STDERR "ERROR: $error_reason\n";
        } else {
                        if (scalar(@content) > 0) {
                                %hash = %{$content[0]};
                        }
        }
        if (defined($hash{gsm})) {
                return $hash{gsm};
        } else {
                return undef;
        }
}



# ELIZA FUNCTIONS HERE

# These are the famous !commands.
sub commands {
   my ($params, $nick, $userhost) = @_;

   # Command is a name only, or a name followed by a space and arguments

   # Commands with arguments
   if ($params =~ /^(.*?)\s(.*?)$/) {
      my $command_name = $1;
      my $command_params = $2;

      # Incremement usage of this function for user
      $cmdusage{$realnick}->{$command_name}++ if $realnick ne "unknown_nick";

      if ($eliza_cmds{ $command_name }) {
         if ($eliza_cmds{$command_name}->{internal}) {
	    my $intcmd = $eliza_cmds{$command_name}->{internal};
	    my $result = &$intcmd($nick, $userhost, $command_params);
            my @lines = split /\n/, $result;
	    my $count = 0;
            foreach my $i (@lines) {
               select undef, undef, undef, $wait_time;
               if ($eliza_cmds{$command_name}->{response} eq "public") {
                  $conn->privmsg( $config{channel}, $i );
               } else {
                  $conn->privmsg( $nick, $i );
               }
            }
	 } elsif ($eliza_cmds{$command_name}->{external}) {
            my $extcmd = $eliza_cmds{$command_name}->{external};
            my $prefix = "";
            $prefix = $eliza_cmds{$command_name}->{external_prefix} if $eliza_cmds{$command_name}->{external_prefix};
            # remove nasty chars
            $command_params =~ s/[;\&]//g;
            open TMP, "$extcmd $command_params |" or return undef;
	    my $count = 0;
            while (<TMP>) {
               chomp;
               select undef, undef, undef, $wait_time;
               my $tmp = $prefix . $_;
               if ($eliza_cmds{$command_name}->{response} eq "public") {
                  $conn->privmsg( $config{channel}, $tmp );
               } else {
                  $conn->privmsg( $nick, $tmp );
               }
            }
            close TMP;
         }
      }
   } elsif ($params =~ /^(\S+)$/) {
      my $command_name = $1;
      my $command_params = "";
   
      # Incremement usage of this function for user
      $cmdusage{$realnick}->{$command_name}++ if $realnick ne "unknown_nick";

      if ($eliza_cmds{$command_name}) {
         if ($eliza_cmds{$command_name}->{internal}) {
	    my $intcmd = $eliza_cmds{$command_name}->{internal};
	    my $result = &$intcmd($nick, $userhost, $command_params);
            my @lines = split /\n/, $result;
            my $count = 0;
            foreach my $i (@lines) {
               select undef, undef, undef, $wait_time;
               if ($eliza_cmds{$command_name}->{response} eq "public") {
                  $conn->privmsg( $config{channel}, $i );
               } else {
                  $conn->privmsg( $nick, $i );
               }
            }
	 } elsif ($eliza_cmds{$command_name}->{external}) {
            my $extcmd = $eliza_cmds{$command_name}->{external};
            my $prefix = "";
            $prefix = $eliza_cmds{$command_name}->{external_prefix} if $eliza_cmds{$command_name}->{external_prefix};
            
            # remove nasty chars
            $command_params =~ s/[;\&]//g;
            open TMP, "$extcmd |" or return undef;
            my $count = 0;
            while (<TMP>) {
               chomp;
               select undef, undef, undef, $wait_time; 
               my $tmp = $prefix . $_;
               if ($eliza_cmds{$command_name}->{response} eq "public") {
                  $conn->privmsg( $config{channel}, $tmp );
               } else {
                  $conn->privmsg( $nick, $tmp );
               }
            }
            close TMP;
         }
      }
   }

   
}






# Divers

sub escape_class {
   my($str, $class) = @_;
   $str =~ s/([$class])/\\$1/g;

   return $str;
}



# Handlers

sub join_channel {
   my ($conn, $event) = @_;
   print( 'We\'re connected!' . "\n" );
   $conn->join( $config{channel} );
   print "We joined $config{channel}\n";
}


sub nickchange {
   my ($conn, $event) = @_;

   my ($nickfrom) = $event->nick;
   my ($nickto) = $event->args;

   my $real_nick_from = &getinfo_from_alias( $nickfrom, 'nick' );
   my $real_nick_to = &getinfo_from_alias( $nickto, 'nick' );

   # Check if this nick is authenticated
   if ($auth{$real_nick_from}->{nick} eq $nickfrom) {
      # If yes, then we try to follow
      if ($real_nick_to eq $real_nick_from) {
         # this means it's the same user, following authentication
	 $auth{$real_nick_from}->{nick} = $nickto;
      } else {
         # else we kill the authentication and inform user
	 delete $auth{$real_nick_from};
	 $conn->privmsg( $nickto, "Your authentication as $real_nick_from as been deleted as destination nick '$nickto' is not you!" );
      }
   }
}



sub quit {
   my ($conn, $event) = @_;

   # Remove authentication when a user quits irc.

   my ($nick) = $event->nick;
   
   my $real_nick = &getinfo_from_alias( $nick, 'nick' );

   if ($auth{$real_nick}) {
      delete $auth{$real_nick};
   }
  
}

sub part {
   my ($conn, $event) = @_;

   # Remove authentication when a user leave channel
   
   my ($nick) = $event->nick;
   
   my $real_nick = &getinfo_from_alias( $nick, 'nick' );

   if ($auth{$real_nick}) {
      delete $auth{$real_nick};
   }
}



sub action {
   my ($conn, $event) = @_;

   my ($msg)  = $event->args;
   my ($nick) = $event->nick;
   my ($dest) = $event->to;
   my ($userhost) = $event->userhost;

   # We save this sentence
   $last_sentences{$nick} = $msg;
}
   


sub message {
   my ($conn, $event) = @_;
   
   my ($msg)  = $event->args;
   my ($nick) = $event->nick;
   my ($dest) = $event->to;
   my ($userhost) = $event->userhost;

   if (!($realname = &getinfo_from_alias( $nick, 'name' ))) {
      $realname = "unknown";
   } 

   if (!($realnick = &getinfo_from_alias( $nick, 'nick' ))) {
      $realnick = "unknown_nick";
   } 

   my $is_not_a_correction = 1;

   # Are we in the configured channel ?
   if ( ($dest eq $config{channel}) || ($dest eq $config{nick}) ) {

      # Let's parse 

      # First, the "!" commands
      if ($msg =~ /^\!(.*)$/) {
         my $params = $1;
         &commands( $params, $nick, $userhost );

      # special commands starting with "$"
      } elsif ($msg =~ /^\$(.*?)$/) {
         my $params = $1;
	 if ($params =~ /^(...)(...):(.*?)$/) {
	    &commands( "currency $1 $2 $3", $nick, $userhost );
	 } else {
	    $conn->privmsg( $config{channel}, "Goodbye $nick ($params)");
	 }

      # special commands starting with "?"
      } elsif ($msg =~ /^\?(.)(.):(.*?)$/) {
         my $s = $1;
         my $d = $2;
         my $t = $3;
         &commands( "trans $s$d $t", $nick, $userhost );

      # special commands starting with "?"
      } elsif ($msg =~ /^=(.*?)$/) {
         &commands( "calc $1", $nick, $userhost );

      # s/// command which is special case
      } elsif ($msg =~ m{^s/((\/|[^/])*)/((\/|[^/])*)/(.*)} ) {
         $is_not_a_correction = undef;
         my $this = $1; 
         my $that = $3;
	 my $flag = $5;

         $this =~ s/([()\[\]{}\\])/\\$1/g;
         $this = &escape_class($this, '?+*{}');

         my $before = $last_sentences{$nick};
         my $lastsaid = $last_sentences{$nick};
	 if ($flag eq "g") {
	    $lastsaid =~ s/$this/$that/g;
         } elsif ($flag eq "i") {
            $lastsaid =~ s/$this/$that/i;
         } elsif ($flag eq "x") {
            $lastsaid =~ s/$this/$that/x;
         } elsif ($flag eq "o") {
            $lastsaid =~ s/$this/$that/o;
         } else {
            $lastsaid =~ s/$this/$that/;
         }

         if ( ($lastsaid ne "") && ($before ne $lastsaid) ) {
            $conn->privmsg( $config{channel}, "$nick corrects: $lastsaid");
         }
         
         $last_sentences{$nick} = $lastsaid;
      }

         
         
      # Check for note for this user
      if ($notes{$realnick}) {
         my $hashref = $notes{$realnick};
         my %tmphash = %$hashref;
         foreach my $i (keys %tmphash) {
            my $hr = $tmphash{$i};
            print "$i : $hr\n";
            my %ha = %$hr;
            foreach my $j (keys %ha){
               print "$j : $ha{$j}\n";
            }
            $conn->privmsg( $nick, "You got this note from " . $tmphash{$i}->{from} . " on " . $i . ":" );
            $conn->privmsg( $nick, "-> " . $tmphash{$i}->{message} );
         }
         # once we displayed the notes, let's remove
	 delete $notes{$realnick};
      }
      

   }

   # We save this sentence, except if it's a s///
   $last_sentences{$nick} = $msg if $is_not_a_correction;
   

   
}


sub read_comm_file {
   open TMP, $config{commfile};
   print <TMP>;
   close TMP;
}



# MAIN
&init;
